@@ -29,11 +29,13 @@ along with Gnuastro. If not, see <http://www.gnu.org/licenses/>.
 #include <stdlib.h>
 #include <gsl/gsl_errno.h>
 
+#include <gnuastro/wcs.h>
+#include <gnuastro/tile.h>
 #include <gnuastro/fits.h>
 #include <gnuastro/threads.h>
-#include <gnuastro/spatialconvolve.h>
+#include <gnuastro/convolve.h>
 
-#include <timing.h>
+#include <gnuastro-internal/timing.h>
 
 #include "main.h"
 #include "convolve.h"
@@ -55,28 +57,26 @@ complextoreal(double *c, size_t size, int action, double **output)
 {
   double *out, *o, *of;
 
-  errno=0;
-  *output=out=malloc(size*sizeof *out);
-  if(out==NULL)
-    error(EXIT_FAILURE, errno, "%lu bytes for out in complextoreal",
-          size*sizeof *out);
+  /* Allocate the space for the real array. */
+  *output=out=gal_data_malloc_array(GAL_TYPE_FLOAT64, size, __func__,
+                                    "output");
 
+  /* Fill the real array with the derived value from the complex array. */
   of=(o=out)+size;
   switch(action)
     {
-    case COMPLEXTOREALSPEC:
+    case COMPLEX_TO_REAL_SPEC:
       do { *o++ = sqrt( *c**c + *(c+1)**(c+1) ); c+=2; } while(o<of);
       break;
-    case COMPLEXTOREALPHASE:
+    case COMPLEX_TO_REAL_PHASE:
       do { *o++ = atan2( *(c+1), *c );           c+=2; } while(o<of);
       break;
-    case COMPLEXTOREALREAL:
+    case COMPLEX_TO_REAL_REAL:
       do { *o++ = *c;                            c+=2; } while(o<of);
       break;
     default:
-      error(EXIT_FAILURE, 0, "a bug! Please contact us at %s so we can "
-            "correct it. For some reason, the action code for "
-            "complextoreal in convolve.c (%d) is not recognized",
+      error(EXIT_FAILURE, 0, "%s: a bug! Please contact us at %s so we can "
+            "correct it. The `action' code %d is not recognized", __func__,
             PACKAGE_BUGREPORT, action);
     }
 }
@@ -118,9 +118,8 @@ complexarraymultiply(double *a, double *b, size_t size)
 
 
 
-/* Divide the elements of the first array by the elements of the
-   second array and put the result into the elements of the first
-   array.
+/* Divide the elements of the first array by the elements of the second
+   array and put the result into the elements of the first array.
 
    (a+ib)/(c+id)=[(a+ib)*(c-id)]/[(c+id)*(c-id)]
                 =(ac-iad+ibc+bd)/(c^2+d^2)
@@ -130,14 +129,14 @@ complexarraymultiply(double *a, double *b, size_t size)
    on the loop.
  */
 void
-complexarraydivide(double *a, double *b, size_t size)
+complexarraydivide(double *a, double *b, size_t size, double minsharpspec)
 {
   double r, *af;
 
   af=a+2*size;
   do
     {
-      if (sqrt(*b**b + *(b+1)**(b+1))>MINGOODDIVSPEC)
+      if (sqrt(*b**b + *(b+1)**(b+1))>minsharpspec)
         {
           r      = ( ( (*a * *b) + (*(a+1) * *(b+1)) )
                      / ( *b * *b + *(b+1) * *(b+1) ) );
@@ -145,6 +144,8 @@ complexarraydivide(double *a, double *b, size_t size)
                      / ( *b * *b + *(b+1) * *(b+1) ) );
           *a=r;
 
+          /* Just as a sanity check (the result should never be larger than
+             one. */
           if(sqrt(*a**a + *(a+1)**(a+1))>1.00001f)
             *a=*(a+1)=0.0f;
         }
@@ -183,55 +184,53 @@ complexarraydivide(double *a, double *b, size_t size)
 /*************      Padding and initializing      *****************/
 /******************************************************************/
 void
-makepaddedcomplex(struct convolveparams *p)
+frequency_make_padded_complex(struct convolveparams *p)
 {
   size_t i, ps0, ps1;
   double *o, *op, *pimg, *pker;
-  float *f, *fp, *input=p->input, *kernel=p->kernel;
-  size_t is0=p->is0, is1=p->is1, ks0=p->ks0, ks1=p->ks1;
+  size_t is0=p->input->dsize[0],  is1=p->input->dsize[1];
+  size_t ks0=p->kernel->dsize[0], ks1=p->kernel->dsize[1];
+  float *f, *ff, *input=p->input->array, *kernel=p->kernel->array;
+
 
   /* Find the sizes of the padded image, note that since the kernel
      sizes are always odd, the extra padding on the input image is
      always going to be an even number (clearly divisable). */
-  ps0=p->ps0 = p->makekernel ? p->is0 : p->is0+p->ks0-1;
-  ps1=p->ps1 = p->makekernel ? p->is1 : p->is1+p->ks1-1;
+  ps0=p->ps0 = p->makekernel ? is0 : is0 + ks0 - 1;
+  ps1=p->ps1 = p->makekernel ? is1 : is1 + ks1 - 1;
+
 
   /* The Discrete Fourier transforms operate faster on even-sized
      arrays. So if the padded sides are not even, make them so: */
   if(ps0%2) ps0=p->ps0=ps0+1;
   if(ps1%2) ps1=p->ps1=ps1+1;
 
+
   /* Allocate the space for the padded input image and fill it. */
-  errno=0;
-  pimg=p->pimg=malloc(2*ps0*ps1*sizeof *pimg);
-  if(pimg==NULL)
-    error(EXIT_FAILURE, errno, "%lu bytes for pimg",
-          2*ps0*ps1*sizeof *pimg);
+  pimg=p->pimg=gal_data_malloc_array(GAL_TYPE_FLOAT64, 2*ps0*ps1, __func__,
+                                     "pimg");
   for(i=0;i<ps0;++i)
     {
       op=(o=pimg+i*2*ps1)+2*ps1; /* pimg is complex.            */
       if(i<is0)
         {
-          fp=(f=input+i*is1)+is1;
-          do {*o++=*f; *o++=0.0f;} while(++f<fp);
+          ff=(f=input+i*is1)+is1;
+          do {*o++=*f; *o++=0.0f;} while(++f<ff);
         }
       do *o++=0.0f; while(o<op);
     }
 
 
   /* Allocate the space for the padded Kernel and fill it. */
-  errno=0;
-  pker=p->pker=malloc(2*ps0*ps1*sizeof *pker);
-  if(pker==NULL)
-    error(EXIT_FAILURE, errno, "%lu bytes for pker",
-          2*ps0*ps1*sizeof *pker);
+  pker=p->pker=gal_data_malloc_array(GAL_TYPE_FLOAT64, 2*ps0*ps1, __func__,
+                                     "pker");
   for(i=0;i<ps0;++i)
     {
       op=(o=pker+i*2*ps1)+2*ps1; /* pker is complex.            */
       if(i<ks0)
         {
-          fp=(f=kernel+i*ks1)+ks1;
-          do {*o++=*f; *o++=0.0f;} while(++f<fp);
+          ff=(f=kernel+i*ks1)+ks1;
+          do {*o++=*f; *o++=0.0f;} while(++f<ff);
         }
       do *o++=0.0f; while(o<op);
     }
@@ -241,48 +240,47 @@ makepaddedcomplex(struct convolveparams *p)
 
 
 
-/*  Remove the padding from the final convolved image and also correct
-    for roundoff errors. Notice that we are putting the pixels in the
-    input image of float type.
+/*  Remove the padding from the final convolved image and also correct for
+    roundoff errors.
 
-    NOTE: The padding to the input image (on the first axis for
-          example) was p->ks0-1. Since p->ks0 is always odd, the
-          padding will alwys be even.  */
+    NOTE: The padding to the input image (on the first axis for example)
+          was `p->kernel->dsize[0]-1'. Since `p->kernel->dsize[0]' is
+          always odd, the padding will always be even.  */
 void
 removepaddingcorrectroundoff(struct convolveparams *p)
 {
   size_t ps1=p->ps1;
-  float *o, *input=p->input;
-  size_t i, hi0, hi1, is0, is1;
-  double *d, *df, *start, *pimg=p->pimg;
+  size_t *isize=p->input->dsize;
+  float *o, *input=p->input->array;
+  double *d, *df, *start, *rpad=p->rpad;
+  size_t i, hi0, hi1, mkwidth=2*p->makekernel-1;
 
-  /* Set all the necessary parameters to crop the desired region. hi0
-     and hi1 are the coordinates of the first pixel in the output
-     image. In the case of deconvolution, if the maximum radius is
-     larger smaller than the input image, we will also only be using
-     region that contains non-zero rows and columns.*/
+  /* Set all the necessary parameters to crop the desired region. hi0 and
+     hi1 are the coordinates of the first pixel in the output image. In the
+     case of deconvolution, if the maximum radius is larger than the input
+     image, we will also only be using region that contains non-zero rows
+     and columns.*/
   if(p->makekernel)
     {
-      hi0 = 2*p->makekernel-1<p->is0 ? p->ps0/2-p->makekernel : 0;
-      hi1 = 2*p->makekernel-1<p->is1 ? p->ps1/2-p->makekernel : 0;
-      p->is0 = 2*p->makekernel-1<p->is0 ? 2*p->makekernel-1 : p->is0;
-      p->is1 = 2*p->makekernel-1<p->is1 ? 2*p->makekernel-1 : p->is1;
+      hi0      = mkwidth < isize[0] ? p->ps0/2-p->makekernel : 0;
+      hi1      = mkwidth < isize[1] ? p->ps1/2-p->makekernel : 0;
+      isize[0] = mkwidth < isize[0] ? 2*p->makekernel-1 : isize[0];
+      isize[1] = mkwidth < isize[1] ? 2*p->makekernel-1 : isize[1];
     }
   else
     {
-      hi0=(p->ks0-1)/2;
-      hi1=(p->ks1-1)/2;
+      hi0 = ( p->kernel->dsize[0] - 1 )/2;
+      hi1 = ( p->kernel->dsize[1] - 1 )/2;
     }
-  is0=p->is0;
-  is1=p->is1;
 
   /* To start with, `start' points to the first pixel in the final
      image: */
-  start=&pimg[hi0*ps1+hi1];
-  for(i=0;i<is0;++i)
+  start=&rpad[hi0*ps1+hi1];
+  for(i=0;i<isize[0];++i)
     {
-      o=&input[i*is1];
-      df=(d=start+i*ps1)+is1;
+      o = &input[ i * isize[1] ];
+
+      df = ( d = start + i * ps1 ) + isize[1];
       do
         *o++ = ( *d<-CONVFLOATINGPOINTERR || *d>CONVFLOATINGPOINTERR )
           ? *d
@@ -310,8 +308,8 @@ fftinitializer(struct convolveparams *p, struct fftonthreadparams **outfp)
   errno=0;
   *outfp=fp=malloc(p->cp.numthreads*sizeof *fp);
   if(fp==NULL)
-    error(EXIT_FAILURE, errno, "%lu bytes in forward2dfft for fp",
-          p->cp.numthreads*sizeof *fp);
+    error(EXIT_FAILURE, errno, "%s: allocating %zu bytes for fp",
+          __func__, p->cp.numthreads*sizeof *fp);
 
   /* Initialize the gsl_fft_wavetable structures (these are thread
      safe): */
@@ -364,19 +362,18 @@ correctdeconvolve(struct convolveparams *p, double **spatial)
 
   /* Check if the image has even sides. */
   if(ps0%2 || ps1%2)
-    error(EXIT_FAILURE, 0, "a bug! Please contact us at %s. In "
-          "correctdeconvolve, the padded image sides are not an "
-          "even number", PACKAGE_BUGREPORT);
+    error(EXIT_FAILURE, 0, "%s: a bug! Please contact us at %s. The padded "
+          "image sides are not an even number", __func__, PACKAGE_BUGREPORT);
 
   /* First convert the complex image to a real image: */
-  complextoreal(p->pimg, ps0*ps1, COMPLEXTOREALSPEC, &s);
+  complextoreal(p->pimg, ps0*ps1, COMPLEX_TO_REAL_SPEC, &s);
 
   /* Allocate the array to keep the new values */
   errno=0;
   n=malloc(ps0*ps1*sizeof *n);
   if(n==NULL)
-    error(EXIT_FAILURE, errno, "%lu bytes for n in correctdeconvolve "
-          "(convolve.c). ", ps0*ps1*sizeof *n);
+    error(EXIT_FAILURE, errno, "%s: allocating %zu bytes for `n'",
+          __func__, ps0*ps1*sizeof *n);
 
 
   /* Put the elements in their proper place: For example in one
@@ -406,7 +403,7 @@ correctdeconvolve(struct convolveparams *p, double **spatial)
           r=sqrt( (ii-ci)*(ii-ci) + (jj-cj)*(jj-cj) );
           sum += n[ii*ps1+jj] = r < p->makekernel ? s[i*ps1+j] : 0;
 
-          /*printf("(%lu, %lu) --> (%lu, %lu)\n", i, j, ii, jj);*/
+          /*printf("(%zu, %zu) --> (%zu, %zu)\n", i, j, ii, jj);*/
         }
     }
 
@@ -484,7 +481,7 @@ onedimensionfft(void *inparam)
      by each other) is stored in p->pimg. So the check below works
      both in the forward and the backward transformation.
   */
-  for(i=0;indexs[i]!=GAL_THREADS_NON_THRD_INDEX;++i)
+  for(i=0; indexs[i]!=GAL_BLANK_SIZE_T; ++i)
     {
       data = ( indexs[i]<maxindex
                ? &pimg[ 2*indexs[i]*indmultip ]   /* *2 because complex. */
@@ -532,10 +529,10 @@ twodimensionfft(struct convolveparams *p, struct fftonthreadparams *fp,
   if(forward1backwardn1==1)       multiple=2;
   else if(forward1backwardn1==-1) multiple=1;
   else
-    error(EXIT_FAILURE, 0, "a bug! In forward2dfft, the value of the "
-          "variable forward1backwardn1 is somehow not 1 or 2, but %d. "
-          "Please contact us at "PACKAGE_BUGREPORT" so we can find the "
-          "cause of the problem and fix it", forward1backwardn1);
+    error(EXIT_FAILURE, 0, "%s: a bug! The value of the variable "
+          "`forward1backwardn1' is %d not 1 or 2. Please contact us at %s "
+          "so we can find the cause of the problem and fix it", __func__,
+          forward1backwardn1, PACKAGE_BUGREPORT);
 
 
   /* ==================== */
@@ -561,7 +558,7 @@ twodimensionfft(struct convolveparams *p, struct fftonthreadparams *fp,
 
       /* Spin off the threads: */
       for(i=0;i<nt;++i)
-        if(indexs[i*thrdcols]!=GAL_THREADS_NON_THRD_INDEX)
+        if(indexs[i*thrdcols]!=GAL_BLANK_SIZE_T)
           {
             fp[i].id=i;
             fp[i].b=&b;
@@ -570,8 +567,8 @@ twodimensionfft(struct convolveparams *p, struct fftonthreadparams *fp,
             fp[i].forward1backwardn1=forward1backwardn1;
             err=pthread_create(&t, &attr, onedimensionfft, &fp[i]);
             if(err)
-              error(EXIT_FAILURE, 0, "can't create thread %lu for rows",
-                    i);
+              error(EXIT_FAILURE, 0, "%s: can't create thread %zu for rows",
+                    __func__, i);
           }
 
       /* Wait for all threads to finish and free the spaces. */
@@ -601,7 +598,7 @@ twodimensionfft(struct convolveparams *p, struct fftonthreadparams *fp,
       else nb=nt+1;
       gal_threads_attr_barrier_init(&attr, &b, nb);
       for(i=0;i<nt;++i)
-        if(indexs[i*thrdcols]!=GAL_THREADS_NON_THRD_INDEX)
+        if(indexs[i*thrdcols]!=GAL_BLANK_SIZE_T)
           {
             fp[i].b=&b;
             fp[i].stride=p->ps1; /* On each column, stride is p->ps1 */
@@ -609,8 +606,8 @@ twodimensionfft(struct convolveparams *p, struct fftonthreadparams *fp,
             fp[i].forward1backwardn1=forward1backwardn1;
             err=pthread_create(&t, &attr, onedimensionfft, &fp[i]);
             if(err)
-              error(EXIT_FAILURE, 0,
-                    "can't create thread %lu for columns", i);
+              error(EXIT_FAILURE, 0, "%s: can't create thread %zu for columns",
+                    __func__, i);
           }
       pthread_barrier_wait(&b);
       pthread_attr_destroy(&attr);
@@ -624,31 +621,40 @@ twodimensionfft(struct convolveparams *p, struct fftonthreadparams *fp,
 
 
 void
-frequencyconvolve(struct convolveparams *p)
+convolve_frequency(struct convolveparams *p)
 {
   double *tmp;
+  size_t dsize[2];
   struct timeval t1;
-  int verb=p->cp.verb;
+  gal_data_t *data=NULL;
   struct fftonthreadparams *fp;
-  char *operation = p->makekernel ? "Divided" : "Multiplied";
 
 
   /* Make the padded arrays. */
-  if(verb) gettimeofday(&t1, NULL);
-  makepaddedcomplex(p);
-  if(verb) gal_timing_report(&t1, "Input and Kernel images padded.", 1);
-  if(p->viewfreqsteps)
+  if(!p->cp.quiet) gettimeofday(&t1, NULL);
+  frequency_make_padded_complex(p);
+  if(!p->cp.quiet)
+    gal_timing_report(&t1, "Input and Kernel images padded.", 1);
+  if(p->checkfreqsteps)
     {
-      complextoreal(p->pimg, p->ps0*p->ps1, COMPLEXTOREALREAL, &tmp);
-      gal_fits_array_to_file(p->up.freqstepsname, "Input padded",
-                             DOUBLE_IMG, tmp, p->ps0, p->ps1, 0,
-                             NULL, NULL, SPACK_STRING);
-      free(tmp);
-      complextoreal(p->pker, p->ps0*p->ps1, COMPLEXTOREALREAL, &tmp);
-      gal_fits_array_to_file(p->up.freqstepsname, "Kernel padded",
-                             DOUBLE_IMG, tmp, p->ps0, p->ps1, 0,
-                             NULL, NULL, SPACK_STRING);
-      free(tmp);
+      /* Prepare the data structure for viewing the steps, note that we
+         don't need the array that is initially made. */
+      dsize[0]=p->ps0; dsize[1]=p->ps1;
+      data=gal_data_alloc(NULL, GAL_TYPE_FLOAT64, 2, dsize, NULL, 0,
+                          p->cp.minmapsize, NULL, NULL, NULL);
+      free(data->array);
+
+      /* Save the padded input image. */
+      complextoreal(p->pimg, p->ps0*p->ps1, COMPLEX_TO_REAL_REAL, &tmp);
+      data->array=tmp; data->name="input padded";
+      gal_fits_img_write(data, p->freqstepsname, NULL, PROGRAM_STRING);
+      free(tmp); data->name=NULL;
+
+      /* Save the padded kernel image. */
+      complextoreal(p->pker, p->ps0*p->ps1, COMPLEX_TO_REAL_REAL, &tmp);
+      data->array=tmp; data->name="kernel padded";
+      gal_fits_img_write(data, p->freqstepsname, NULL, PROGRAM_STRING);
+      free(tmp); data->name=NULL;
     }
 
 
@@ -657,73 +663,72 @@ frequencyconvolve(struct convolveparams *p)
 
 
   /* Forward 2D FFT on each image. */
-  if(verb) gettimeofday(&t1, NULL);
+  if(!p->cp.quiet) gettimeofday(&t1, NULL);
   twodimensionfft(p, fp, 1);
-  if(verb)
+  if(!p->cp.quiet)
     gal_timing_report(&t1, "Images converted to frequency domain.", 1);
-  if(p->viewfreqsteps)
+  if(p->checkfreqsteps)
     {
-      complextoreal(p->pimg, p->ps0*p->ps1, COMPLEXTOREALSPEC, &tmp);
-      gal_fits_array_to_file(p->up.freqstepsname, "Input transform",
-                             DOUBLE_IMG, tmp, p->ps0, p->ps1, 0,
-                             NULL, NULL, SPACK_STRING);
-      free(tmp);
-      complextoreal(p->pker, p->ps0*p->ps1, COMPLEXTOREALSPEC, &tmp);
-      gal_fits_array_to_file(p->up.freqstepsname, "Kernel transform",
-                             DOUBLE_IMG, tmp, p->ps0, p->ps1, 0,
-                             NULL, NULL, SPACK_STRING);
-      free(tmp);
+      complextoreal(p->pimg, p->ps0*p->ps1, COMPLEX_TO_REAL_SPEC, &tmp);
+      data->array=tmp; data->name="input transformed";
+      gal_fits_img_write(data, p->freqstepsname, NULL, PROGRAM_STRING);
+      free(tmp); data->name=NULL;
+
+      complextoreal(p->pker, p->ps0*p->ps1, COMPLEX_TO_REAL_SPEC, &tmp);
+      data->array=tmp; data->name="kernel transformed";
+      gal_fits_img_write(data, p->freqstepsname, NULL, PROGRAM_STRING);
+      free(tmp); data->name=NULL;
     }
 
   /* Multiply or divide the two arrays and save them in the output.*/
-  if(verb) gettimeofday(&t1, NULL);
+  if(!p->cp.quiet) gettimeofday(&t1, NULL);
   if(p->makekernel)
     {
-      complexarraydivide(p->pimg, p->pker, p->ps0*p->ps1);
-      if(verb)
+      complexarraydivide(p->pimg, p->pker, p->ps0*p->ps1, p->minsharpspec);
+      if(!p->cp.quiet)
         gal_timing_report(&t1, "Divided in the frequency domain.", 1);
     }
   else
     {
       complexarraymultiply(p->pimg, p->pker, p->ps0*p->ps1);
-      if(verb)
+      if(!p->cp.quiet)
         gal_timing_report(&t1, "Multiplied in the frequency domain.", 1);
     }
-  if(p->viewfreqsteps)
+  if(p->checkfreqsteps)
     {
-      complextoreal(p->pimg, p->ps0*p->ps1, COMPLEXTOREALSPEC, &tmp);
-      gal_fits_array_to_file(p->up.freqstepsname, operation,
-                             DOUBLE_IMG, tmp, p->ps0, p->ps1, 0,
-                             NULL, NULL, SPACK_STRING);
-      free(tmp);
+      complextoreal(p->pimg, p->ps0*p->ps1, COMPLEX_TO_REAL_SPEC, &tmp);
+      data->array=tmp; data->name=p->makekernel ? "Divided" : "Multiplied";
+      gal_fits_img_write(data, p->freqstepsname, NULL, PROGRAM_STRING);
+      free(tmp); data->name=NULL;
     }
 
   /* Forward (in practice inverse) 2D FFT on each image. */
-  if(verb) gettimeofday(&t1, NULL);
+  if(!p->cp.quiet) gettimeofday(&t1, NULL);
   twodimensionfft(p, fp, -1);
-  if(p->makekernel) correctdeconvolve(p, &tmp);
-  else complextoreal(p->pimg, p->ps0*p->ps1, COMPLEXTOREALREAL, &tmp);
-  if(verb)
+  if(p->makekernel)
+    correctdeconvolve(p, &p->rpad);
+  else
+    complextoreal(p->pimg, p->ps0*p->ps1, COMPLEX_TO_REAL_REAL, &p->rpad);
+  if(!p->cp.quiet)
     gal_timing_report(&t1, "Converted back to the spatial domain.", 1);
-  if(p->viewfreqsteps)
+  if(p->checkfreqsteps)
     {
-      gal_fits_array_to_file(p->up.freqstepsname, "Spatial",
-                             DOUBLE_IMG, tmp, p->ps0, p->ps1, 0,
-                             NULL, NULL, SPACK_STRING);
-      free(tmp);
+      data->array=p->rpad; data->name="padded output";
+      gal_fits_img_write(data, p->freqstepsname, NULL, PROGRAM_STRING);
+      data->name=NULL; data->array=NULL;
     }
 
   /* Free the padded arrays (they are no longer needed) and put the
      converted array (that is real, not complex) in p->pimg. */
+  gal_data_free(data);
   free(p->pimg);
   free(p->pker);
-  p->pimg=tmp;
 
   /* Crop out the center, numbers smaller than 10^{-17} are errors,
      remove them. */
-  if(verb) gettimeofday(&t1, NULL);
+  if(!p->cp.quiet) gettimeofday(&t1, NULL);
   removepaddingcorrectroundoff(p);
-  if(verb) gal_timing_report(&t1, "Padded parts removed.", 1);
+  if(!p->cp.quiet) gal_timing_report(&t1, "Padded parts removed.", 1);
 
 
   /* Free all the allocated space. */
@@ -754,44 +759,41 @@ frequencyconvolve(struct convolveparams *p)
 void
 convolve(struct convolveparams *p)
 {
-  float *convolved;
-  long *meshindexs;
-  struct gal_mesh_params *mp=&p->mp;
+  gal_data_t *out, *check;
+  struct gal_options_common_params *cp=&p->cp;
+
 
   /* Do the convolution. */
-  if(p->spatial)
+  if(p->domain==CONVOLVE_DOMAIN_SPATIAL)
     {
-      /* Prepare the mesh structure: */
-      mp->img=p->input;      mp->s0=p->is0;      mp->s1=p->is1;
-      mp->kernel=p->kernel;  mp->ks0=p->ks0;     mp->ks1=p->ks1;
-      mp->numthreads=p->cp.numthreads;
-      gal_mesh_make_mesh(mp);
-      if(p->meshname)
+      /* Prepare the mesh structure. */
+      gal_tile_full_two_layers(p->input, &cp->tl);
+
+      /* Save the tile IDs if they are requested. */
+      if(cp->tl.tilecheckname)
         {
-          gal_mesh_check_mesh_id(mp, &meshindexs);
-          gal_fits_array_to_file(p->meshname, "Input", FLOAT_IMG,
-                                 p->mp.img, mp->s0, mp->s1,
-                                 p->anyblank, p->wcs, NULL,
-                                          SPACK_STRING);
-          gal_fits_array_to_file(p->meshname, "MeshIndexs", LONG_IMG,
-                                 meshindexs, mp->s0, mp->s1, 0, p->wcs,
-                                 NULL, SPACK_STRING);
-          free(meshindexs);
+          check=gal_tile_block_check_tiles(cp->tl.tiles);
+          gal_fits_img_write(check, cp->tl.tilecheckname, NULL, PROGRAM_NAME);
+          gal_data_free(check);
         }
 
-      /* Do the spatial convolution on the mesh: */
-      gal_mesh_spatial_convolve_on_mesh(mp, &convolved);
+      /* Do the spatial convolution. One of the main reason someone would
+         want to do spatial domain convolution with this Convolve program
+         is edge correction. So by default we assume it and will only
+         ignore it if the user asks.*/
+      out=gal_convolve_spatial(cp->tl.tiles, p->kernel, cp->numthreads,
+                               !p->noedgecorrection, cp->tl.workoverch);
 
-      /* Replace the input image array with the convolved array: */
-      free(p->input);
-      p->input=convolved;
+      /* Clean up: free the actual input and replace it's pointer with the
+         convolved dataset to save as output. */
+      gal_tile_full_free_contents(&cp->tl);
+      gal_data_free(p->input);
+      p->input=out;
     }
   else
-    frequencyconvolve(p);
+    convolve_frequency(p);
 
-  /* Save the output (which is in p->input) array. Note that p->input
-     will be freed in ui.c. */
-  gal_fits_array_to_file(p->cp.output, "Convolved", FLOAT_IMG,
-                         p->input, p->is0, p->is1, p->anyblank, p->wcs,
-                         NULL, SPACK_STRING);
+  /* Save the output (which is in p->input) array. */
+  gal_fits_img_write_to_type(p->input, cp->output, NULL, PROGRAM_STRING,
+                             cp->type);
 }
